#!/usr/bin/perl
#
# $Id$
#
# (C) IBCL Informatik-Buero e.U. 2010
# (C) ITEG IT-Engineers GmbH 2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;

use HTTP::Request;
use HTTP::Request::Common;
use List::MoreUtils qw/ uniq /;
use LWP::UserAgent;

require "/usr/lib/pba/pba-functions.pl";


sub prepareGetopts() {

  $main::helpMessage = '
 Syntax:
  '.$main::scriptName.' CommandOption [Options]
  '.$main::scriptName.' --help            Prints this help text and exits.
  '.$main::scriptName.' --version         Prints the version info and exits.
 
 CommandOptions are:
  -A               Automatic mode, build eventual incoming .dsc/.srpm files.
  -B directory     Automatic mode, build eventual incomings in directory
  -C distribution  Create a pbabuilder tarball.
  -E distribution  Enter a pbabuilder tarball (for manual changes).
  -L distribution  Login to a pbabuilder tarball (for manual research).
  -U distribution  Update a pbabuilder tarball (i.e. the vendor packages).
                   Also installs packages needed for all builds.
  -T               Test operation, internal use only.
  -X dist -x x.sh  Execute x.sh in a pbabuilder tarball (for automated changes).
  --help           Prints this help text and exits.
  --version        Prints the version info and exits.
 
 Options are:
  -a arch          deb only: Target architecture is arch, not amd64. e.g. armhf
  -b binPkgsDir    directory to cp results to, do not feed any repository
  -c configdir     Read configdir/* rather than /etc/pba/*
  -d               Debug mode, logs everything.
  -q               Quiet mode, supresses informative messages.
  -v               Verbose verboser, logs more.
 
 For some useful shell loops take a look at ReadMe.txt, usually at
 /usr/share/doc/pba-common/ReadMe.txt
';

  $main::optionList = "AB:C:E:L:TU:X:a:b:c:dqvx:";
  $main::role = "builder";
  $main::logDir = "/var/log/pba-builder";
  (-d $main::logDir) || cleanupAndDie("FATAL: Log directory does not exist: ".$main::logDir."\n");
  $main::calledWD = pwd();
  
}


sub initBuildFunctions() {

  $main::configParams{"Architectures"} = "amd64";
  $main::configParams{"BuildReportsFrom"} = "";
  $main::configParams{"BuildReportsCc"} = "";
  $main::configParams{"DefaultDistribution"} = "stretch";
  $main::configParams{"HttpProxy"} = "";
  $main::configParams{"HttpSProxy"} = "";
  $main::configParams{"GPGHome"} = "";
  $main::configParams{"GPGKeyId"} = "";
  $main::configParams{"GPGOptions"} = "";
  $main::configParams{"IncomingDirs"} = "";
  $main::configParams{"MaxParallelBuilds"} = 1;
  $main::configParams{"ResultsDir"} = "";
  $main::configParams{"SMTPHost"} = "localhost";
  $main::configParams{"SMTPPort"} = 25;
  $main::configParams{"SMTPUser"} = "";
  $main::configParams{"SMTPPassword"} = "";

}


sub checkBuildConfig() {

  chdir $main::calledWD || cleanupAndDie("FATAL: Cannot chdir back to ".$main::calledWD."\n");

  # check config
  $main::incomingDirs = simplifyList($main::configParams{"IncomingDirs"});
  ($main::incomingDirs eq "") && cleanupAndDie("FATAL: No IncomingDirs specified in ".$main::configFile."\n");

  $main::resultsDir = $main::configParams{"ResultsDir"};
  ($main::resultsDir eq "") && cleanupAndDie("FATAL: No ResultsDir specified in ".$main::configFile."\n");

  $main::gpgHome = $main::configParams{"GPGHome"};
  $main::gpgKeyId = $main::configParams{"GPGKeyId"};
  (length($main::configParams{"GPGOptions"}) > 0) && cleanupAndDie("FATAL: Obsolete option GPGOptions found in ".$main::configFile.", please migrate to GPGHome, GPGKeyId.\n");

  $main::maxParallelBuilds = scalar($main::configParams{"MaxParallelBuilds"});
  ($main::maxParallelBuilds < 1) && cleanupAndDie("FATAL: MaxParallelBuilds (".$main::maxParallelBuilds.") too low (<0) specified in ".$main::configFile."\n");
  ($main::maxParallelBuilds > 16) && cleanupAndDie("FATAL: MaxParallelBuilds (".$main::maxParallelBuilds.") too high (>16) specified in ".$main::configFile."\n");
  $main::ignoreMaxParallelBuilds = 0;

  $main::targetArchitectures = $main::configParams{"Architectures"};
  $main::buildReportsCc = $main::configParams{"BuildReportsCc"};
  $main::buildReportsFrom = $main::configParams{"BuildReportsFrom"};
  ($main::buildReportsFrom eq "") && cleanupAndDie("FATAL: No BuildReportsFrom specified in ".$main::configFile."\n");
  $main::defaultDistribution = $main::configParams{"DefaultDistribution"};

  $main::smtpHost = $main::configParams{"SMTPHost"};
  $main::smtpPort = $main::configParams{"SMTPPort"};
  $main::smtpUser = $main::configParams{"SMTPUser"};
  $main::smtpPassword = $main::configParams{"SMTPPassword"};
  $main::smtpEhlo = "localhost.localdomain";
  if ( -f "/etc/mailname") {
    $main::smtpEhlo = readCompleteFile("/etc/mailname");
  }

  foreach my $repository (keys(%main::repositoryRoots)) {
     ($main::repositoryFeedHostDirs{$repository} eq "") && cleanupAndDie("FATAL: No FeedHostDir specified for repository ".$repository."\n");
  }

  # scan command and options
  $main::targetDistribution = "";
  my $commandOptions = 0;
  if ($main::opts{"A"} ne "") {
    $main::buildOperation = "auto-build";
    $commandOptions++;
  }
  if ($main::opts{"B"} ne "") {
    $main::buildOperation = "auto-build-directory";
    $main::incomingDirs = $main::opts{"B"};
    if ($main::incomingDirs !~ /^\//) {
      $main::incomingDirs = pwd()."/".$main::opts{"B"};
    }
    if (! -d $main::incomingDirs) {
      cleanupAndDie("FATAL: Directory specified after -B does not exist: ".$main::incomingDirs."\n");
    }
    $commandOptions++;
    $main::ignoreMaxParallelBuilds = 1;
  }
  if ($main::opts{"C"} ne "") {
    $main::buildOperation = "create-pbabuilder";
    $main::targetDistribution = $main::opts{"C"};
    $commandOptions++;
  }
  if ($main::opts{"E"} ne "") {
    $main::buildOperation = "enter-pbabuilder";
    $main::targetDistribution = $main::opts{"E"};
    $commandOptions++;
  }
  if ($main::opts{"L"} ne "") {
    $main::buildOperation = "login-pbabuilder";
    $main::targetDistribution = $main::opts{"L"};
    $commandOptions++;
  }
  if ($main::opts{"T"} ne "") {
    $main::buildOperation = "test";
    $commandOptions++;
  }
  if ($main::opts{"U"} ne "") {
    $main::buildOperation = "update-pbabuilder";
    $main::targetDistribution = $main::opts{"U"};
    $commandOptions++;
  }
  if ($main::opts{"X"} ne "") {
    $main::buildOperation = "execute-in-pbabuilder";
    $main::targetDistribution = $main::opts{"X"};
    $main::scriptToExecute = $main::opts{"x"};
    ($main::scriptToExecute eq "") && cleanupAndDie("FATAL: No or empty -x option after -X.\n");
    (-x $main::scriptToExecute) || cleanupAndDie("FATAL: No such executable file: ".$main::scriptToExecute.".\npwd=".pwd()."\n");
    ($main::scriptToExecute =~ /\//) && cleanupAndDie("FATAL: Script not in current dir: ".$main::scriptToExecute.".\n");
    $commandOptions++;
  }
  $main::pbabuilderArchitecture = myArchitecture();
  $main::targetArchitecture = $main::opts{"a"};
  if ($main::targetArchitecture eq "") {
    $main::targetArchitecture = $main::pbabuilderArchitecture;
  } else {
    my $hostArchitecture = $main::pbabuilderArchitecture;
    #if ($hostArchitecture ne $main::targetArchitecture) {
    #  if ($main::targetArchitecture eq "i386") {
    #    cleanupAndDie("FATAL: Support for i386 crossbuilds has been dropped as of PBA 1.6, see PBA-24\n");
    #  }
    #  if ($main::targetArchitecture ne "armhf") {
    #    cleanupAndDie("FATAL: Crossbuilds only support armhf, not for ".$main::targetArchitecture."\n");
    #  }
    #}
  }

  $main::localTargetDirectory = "";
  if ($main::opts{"b"} ne "") {
    $main::localTargetDirectory = $main::opts{"b"};
    if ($main::localTargetDirectory !~ /^\//) {
      $main::localTargetDirectory = pwd()."/".$main::opts{"b"};
    }
    checkDirectory($main::localTargetDirectory);
  }

  # check command and options
  if ($commandOptions == 0) {
    print "FATAL: No command specified!\n";
    print $main::helpMessage;
    exit 1;
  }
  if ($commandOptions > 1) {
    cleanupAndDie("FATAL: Conflicting Commands (only one of -A, -B, -C, -I, -L, -U, -X, allowed)!\n");
  }
  if ( ($commandOptions == 1) and ($main::buildOperation =~ /pbabuilder/) ) {
    ($main::targetDistribution eq "") && cleanupAndDie("FATAL: No distribution specified for pbabuilder-operation\n");
    ($main::distributionLongNames{$main::targetDistribution} eq "") && cleanupAndDie("FATAL: Unknown Distribution for pbabuilder-operation:".$main::targetDistribution."\n");
    ($main::verbosity > 0) && print "Target distribution: ".$main::targetDistribution.", ".$main::distributionLongNames{$main::targetDistribution}."\n";
    ($main::distributionMirrors{$main::targetDistribution} =~ /^(http|ftp):\/\//) || cleanupAndDie("FATAL: Distribution mirror does not look valid: ".$main::distributionMirrors{$main::targetDistribution}."\n");
    ($main::verbosity > 0) && print "Target architecture: ".$main::targetArchitecture."\n";
  }

  # signature options
  if ($main::gpgKeyId eq "") {
    $main::dpkgGpgOptions = "-us -uc";
    ($main::verbosity > 0) && print "No GPG key configured, not signing deb packages.\n";
  } else {
    if ($main::opts{"u"}) {
      $main::dpkgGpgOptions = "-us -uc";
      ($main::verbosity > 0) && print "Creating unsigned deb packages despite GPG key, due to -u.\n";
    } else {
      $main::dpkgGpgOptions = "-k".$main::gpgKeyId;
      ($main::verbosity > 0) && print "Using GPG key with ID ".$main::gpgKeyId." to sign deb packages.\n";
    }
  }

  # debug logging
  if ($main::verbosity > 2) {
    print "Incoming directories: ".$main::incomingDirs."\n";
    print "BuildReportsFrom: ".$main::buildReportsFrom."\n";
    print "Results directory: ".$main::resultsDir."\n";
    print "GPG Home: '".$main::gpgHome."'\n";
    print "GPG KeyId: '".$main::gpgKeyId."'\n";
    print "Max parallel builds: ".$main::maxParallelBuilds."\n";
  }

}


sub reportAndDie ( $ $ $ ) {
  # $1 = full build report
  my $report = shift;
  # $2 = maintainer
  my $maintainer = shift;
  # $3 = die message
  my $dieMessage = shift;

  $report .= "\nFATAL: ".$dieMessage."\n";

  my $attachments = "";
  report ( $report, $maintainer, $attachments );

  cleanupAndDie("FATAL: ".$dieMessage."\n".$report."\n");
}


sub report ( $ $ $ ) {
  # $1 = full build report
  my $report = shift;
  # $2 = maintainer
  my $maintainers = shift;
  my @maintainersNamesAddresses = split /\s*,\s*/, $maintainers;
  my $toRecipient = shift @maintainersNamesAddresses;
  # $3 = list of log files to attach
  my $filesToAttach = shift;

  my $maintainer = split
  # use first line of $report as subject
  my $subject = $report;
  $subject =~ s/\n.*//gm;

  # add report cc
  my $ccRecipients = $main::buildReportsCc;
  if (scalar(@maintainersNamesAddresses) != 0) {
    if ($ccRecipients eq "") {
      $ccRecipients = join(", ", @maintainersNamesAddresses);
    } else {
      $ccRecipients .= ", ".join(", ", @maintainersNamesAddresses);
    }
  }
  # make unique
  my %uniqueCcRecipients=();
  foreach my $ccRecipient (split /\s*,\s*/, $ccRecipients) {
    next if ($ccRecipient eq $toRecipient);
    $uniqueCcRecipients{$ccRecipient}=$ccRecipient;
  }
  my $uniqueCommaCcRecipients = join(", ", sort(keys(%uniqueCcRecipients)));

  if ($main::verbosity > 2) {
    print "DEBUG: report(): maintainers='".$maintainers."'\n";
    print "DEBUG: report(): maintainersNamesAddresses='".join("','", @maintainersNamesAddresses)."'\n";
    print "DEBUG: report(): toRecipients='".$toRecipient."'\n";
    print "DEBUG: report(): ccRecipients='".$ccRecipients."'\n";
    print "DEBUG: report(): uniqueCcRecipients='".$uniqueCommaCcRecipients."'\n";
    print "DEBUG: report(): subject='".$subject."'\n";
  }

  # mail ...
  sendEMail( $main::buildReportsFrom, $toRecipient, $uniqueCommaCcRecipients, $subject, $report, $filesToAttach );
  
}


1;

