#!/usr/bin/perl
#
# $Id$
#
# (C) ITEG IT-Engineers GmbH 2015-2022
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;

use HTTP::Request;
use HTTP::Request::Common;
use List::MoreUtils qw/ uniq /;
use LWP::UserAgent;

require "/usr/lib/pba/pba-functions.pl";
require "/usr/lib/pba/build-functions.pl";


sub initDebBuildFunctions() {

  # nothing deb specific yet

}


sub checkDebBuildConfig() {

  # nothing deb specific yet

}


sub scanForIncomingDscFiles() {
  #%main::dscFilesToBuildByTimestamp = ();
  $main::dscFileToBuild = "";
  my $oldestTimestamp = time() + 1;
  my @dirsToScan = split / +/, $main::incomingDirs;
  foreach my $dirToScan (@dirsToScan) {
    ($main::verbosity > 2) && print "Scanning incoming directory ".$dirToScan."\n";
    chdir $dirToScan || cleanupAndDie("FATAL: Cannot chdir to ".$dirToScan."\n");
    my @files = glob("*.dsc");
    foreach my $file (@files) {
      next if (! -f $file); # skip dirs
      ($main::verbosity > 2) && print "Found source package: ".$file."\n";
      my $fileTimestamp = (stat $file)[9];
      if ($fileTimestamp lt $oldestTimestamp) { 
        $main::dscFileToBuild = $dirToScan."/".$file;
        $oldestTimestamp = $fileTimestamp;
      }
    }
  }
  ( ($main::verbosity > 2) and ($main::dscFileToBuild ne "") ) && print "Oldest found source package is: ".$main::dscFileToBuild."\n";
}


sub buildDscFile( $ ) {

  my $dscImportFile = shift;
  ($dscImportFile eq "") && cleanupAndDie("FATAL: buildDscFile called w/o file parameter.\n");

  my @dscImportDirParts = split /\//, $dscImportFile;
  my $dscFile = pop @dscImportDirParts;
  my $dscDir = join '/', @dscImportDirParts;
  #print "dscDirName = $dscDir\n";
  #print "dscFileName = $dscFile\n";

  my $workDir = "/var/spool/pba-builder/work/".$$;
  checkDirectory( $workDir, "0755" );
  push @main::dirsToCleanup, $workDir;
  push @main::dirsToCleanup, "/tmp/pba-builder_".$$;
  chdir $workDir || cleanupAndDie("FATAL: Cannot chdir to ".$workDir."\n");

  # move .dsc file and childs out of brothers' way

  executeOrDie("/bin/mv ".$dscDir."/".$dscFile." ".$workDir."/");

  my $buildHeadlines = "Build report for ".$dscFile."\n\n";
  my $buildReport = "Full build report for ".$dscFile."\n\n";

  # parse changes file
  my $changesFile = $dscFile;
  $changesFile =~ s/\.dsc$/_source.changes/;
  my $changesText = "";
  my $inFilesSection = 0;
  my $inDscSignature = 0;
  my $changedBy = "";
  my $maintainer = "";
  my $maintainers = "";
  my $changesVersion = "";
  executeOrDie("/bin/mv ".$dscDir."/".$changesFile." ".$workDir."/");
  open (CHANGES, "<".$changesFile) || cleanupAndDie("FATAL: Cannot read ".$changesFile."\n");
  while ( my $curLine = <CHANGES> ) {
    chomp $curLine;
    if ($curLine =~ /PGP SIGNATURE/i) {
      $inDscSignature = 1;
      next;
    }
    next if ($inDscSignature);
    next if ($curLine =~ /^(----|Hash: |\s*$)/i);
    $changesText .= $curLine."\n";
    next if ( $curLine !~ /:/ );
    my ( $pn, $pv ) = split /: */, $curLine, 2;
    if ($pn =~ /^Changed-By$/ ) {
      $changedBy = $pv;
      next;
    }
    if ($pn =~ /^Version$/ ) {
      $changesVersion = $pv;
      next;
    }
  }
  close CHANGES; 
  $maintainers = $changedBy;

  # check dsc signature
  my $gpgCmd = "LANG='C' gpg --batch";
  if (length($main::gpgHome) > 0) {
    $gpgCmd .= " --homedir ".$main::gpgHome;
  }
  $gpgCmd .= " --verify ".$workDir."/".$dscFile;
  print "Checking dsc signature: ".$gpgCmd."\n";
  if (execute($gpgCmd)) {
    $buildHeadlines .= "GOOD: Signature checked OK: ".$dscFile."\n";
    $buildReport .= "GOOD: Signature checked OK: ".$dscFile."\n";
  } else {
    #$buildHeadlines .= "FATAL: Signature check failed on ".$dscFile."\n";
    #$buildReport .= "FATAL: Signature check failed on ".$dscFile."\n";
    #reportAndDie($buildReport, $changedBy, "Signature check failed on ".$dscFile."\n");
    $buildHeadlines .= "WARNING: Signature check failed on ".$dscFile."\n";
    $buildReport .= "WARNING: Signature check failed on ".$dscFile."\n";
  }

  # parse dsc file 
  if ($main::verbosity > 0) { $buildReport .= "Analysing grabbed ".$dscDir."/".$dscFile."\n"; }
  my $dscArchitecture = "";
  my $dscBuildDistribution = "";
  my $dscCategory = "";
  my @dscDistributions = ();
  my $dscGeneration = "";
  my $dscMaintainer = "";
  my @dscRepositories = "";
  my $inFilesSection = 0;
  my @dscChildFiles = ();
  #my $dscText = "";
  my $inDscSignature = 0;
  my $dscVersion = ""; # distinct from $dscFormat
  my $dscAptSources = "";
  my @dscTrustedKeyURLs = ();
  my $dscSrcName = "unnamed_src";
  open (DSC, "<".$dscFile) || cleanupAndDie("FATAL: Cannot read ".pwd()."/".$dscFile."\n");
  while ( my $curLine = <DSC> ) {
    chomp $curLine;
    if ($curLine =~ /PGP SIGNATURE/i) {
      $inDscSignature = 1;
      next;
    }
    next if ($inDscSignature);
    next if ($curLine =~ /^(----|Hash: |\s*$)/i);
    #$dscText .= $curLine."\n";
    if ($inFilesSection > 0) {
      if ( $curLine =~ /^\s+\w+\s+\w+\s+/ ) {
        my ($hash, $nn, $childFilename) = split / /, simplifyList($curLine);
        if ($main::verbosity > 1) { $buildReport .= "Grabbing child file: ".$dscDir."/".$childFilename."\n"; }
        if (not -f $dscDir."/".$childFilename) {
          reportAndDie($buildReport, $maintainers, "Child file missing: ".$dscDir."/".$childFilename);
        }
        executeOrDie("/bin/mv ".$dscDir."/".$childFilename." ".$workDir."/");
        push @dscChildFiles, $childFilename; 
        next;
      } else {
        $inFilesSection = 0;
      }
    }
    next if ( $curLine !~ /:/ );
    my ( $pn, $pv ) = split /: */, $curLine, 2;
    if ($pn =~ /^Version$/i) {
      $dscVersion = $pv;
      next;
    }
    if ($pn =~ /^Files$/i) {
      $inFilesSection = 1;
      next;
    }
    if ($pn =~ /^Architecture$/i) {
      $dscArchitecture = $pv;
      next;
    }
    if ($pn =~ /^Maintainer$/i) {
      $maintainer = $pv;
      if ($maintainer ne $changedBy) {
        $maintainers .= ", ".$pv
      }
      next;
    }
    # let's not bother Uploaders from packages we tweak or re-package
    #if ($pn =~ /^Uploaders$/i) {
    #  if ($pv ne $changedBy) {
    #    $maintainers .= ", ".$pv
    #  }
    #  next;
    #}
    if ($pn =~ /^PBA-Build-Distribution$/i) {
      $dscBuildDistribution = $pv;
      next;
    }
    if ($pn =~ /^PBA-Category$/i) {
      $dscCategory = $pv;
      next;
    }
    if ($pn =~ /^PBA-BuildReportCc/i) {
      if ($pv ne $changedBy) {
        $maintainers .= ", ".$pv
      }
      next;
    }
    if ($pn =~ /^PBA-Distributions$/i) {
      @dscDistributions = split / /, simplifyList($pv);
      next;
    }
    if ($pn =~ /^PBA-Generation$/i) {
      $dscGeneration = $pv;
      next;
    }
    if ($pn =~ /^PBA-Repositories$/i) {
      @dscRepositories = split / /, simplifyList($pv);
      next;
    }
    if ($pn =~ /^PBA-Build-Sources?-([a-z0-9_-]+)/i) {
      $dscAptSources .= "# from ".$&.":\n".$pv."\n";
      if ($dscVersion =~ /nightly/) {
        #   deb http://deb.clazzes.org/$family $distribution-fancylibs-3nightly main
        # in addition to
        #   deb http://deb.clazzes.org/$family $distribution-fancylibs-3 main
        my @asParts = split / +/, $pv;
        $dscAptSources .= $asParts[0]." ".$asParts[1]." ".$asParts[2]."nightly ".$asParts[3]."\n";
        #$dscAptSources .= $asParts[0]." ".$asParts[1]." ".$asParts[2]."beta ".$asParts[3]."\n";
      }
      if ($dscVersion =~ /beta|nightly/) {
        #   deb http://deb.clazzes.org/$family $distribution-fancylibs-3beta main
        # in addition to
        #   deb http://deb.clazzes.org/$family $distribution-fancylibs-3 main
        my @asParts = split / +/, $pv;
        $dscAptSources .= $asParts[0]." ".$asParts[1]." ".$asParts[2]."beta ".$asParts[3]."\n";
      }
      next;
    }
    if ($pn =~ /^PBA-TrustedKey-URL-([a-z0-9_-]+)/i) {
      push @dscTrustedKeyURLs, $pv;
      next;
    }
    if ($pn =~ /^Source$/i) {
      $dscSrcName = $pv;
      next;
    }
  }
  close DSC;
  ($dscArchitecture eq "") && reportAndDie($buildReport, $maintainers, "No Architecture specified in ".$dscFile."\n");
  ($dscCategory eq "") && reportAndDie($buildReport, $maintainers, "No PBA-Category specified in ".$dscFile."\n");
  (scalar(@dscDistributions) < 1) && reportAndDie($buildReport, $maintainers, "No PBA-Distributions specified in ".$dscFile."\n");
  ($dscGeneration eq "") && reportAndDie($buildReport, $maintainers, "No PBA-Generation specified in ".$dscFile."\n");
  foreach my $dscDistribution (@dscDistributions) {
    (defined $main::distributionLongNames{$dscDistribution}) || reportAndDie($buildReport, $maintainers, "Unknown distribution ".$dscDistribution." in ".$dscFile."\n");
  }
  (scalar(@dscRepositories) == 0) && reportAndDie($buildReport, $maintainers, "No PBA-Repositories specified in neither ".$dscFile." nor for category ".$dscCategory."\n");
  foreach my $dscRepository (@dscRepositories) {
    (defined $main::repositoryLongNames{$dscRepository}) || reportAndDie($buildReport, $maintainers, "Unknown repository ".$dscRepository." in ".$dscFile."\n");
  }

  my $buildAptRepository = $dscRepositories[0];

  if ($main::verbosity > 1) { $buildReport .= "Package file seems ok: ".$dscFile."\n"; }
  
  # future: check if others must be built first?

  # $dscArchitecture can i.e. be:
  # "any"
  # "all"
  # "amd64"
  # "any all"
  # "amd64 all"
  # "all any"
  # "all amd64"
  my $indepArchitecture = "";   # "all" or empty
  my @archArchitectures = (); # non-indep architectures
  my @srcArchitectures = split / /, simplifyList($dscArchitecture);
  foreach my $srcArchitecture (@srcArchitectures) {
    if ($main::verbosity > 3) { $buildReport .= "Analysing source package architecture (".$srcArchitecture.").\n"; }
    if ($srcArchitecture eq "any") {
      if ($main::verbosity > 2) { $buildReport .= "Found source package architecture 'any', building for all my targetArchitectures (".$main::targetArchitectures.").\n"; }
      push @archArchitectures, split / /, simplifyList($main::targetArchitectures);
      next;
    }
    if ($srcArchitecture eq "all") {
      if ($main::verbosity > 2) { $buildReport .= "Found soure package architecture 'all', remebering to build for default arch if no non-indep architectures come up.\n"; }
      $indepArchitecture = "all";
      next;
    }
    if ($main::verbosity > 2) { $buildReport .= "Found source package architecture '".$srcArchitecture."', treating this as non-indep architecture.\n"; }
    push @archArchitectures, $srcArchitecture;
  }
  my $nrArchArchitectures = scalar(@archArchitectures);
  my @dscArchitectures = ();
  if ($nrArchArchitectures > 0) {
    @dscArchitectures = @archArchitectures;
    if ($main::verbosity > 2) { $buildReport .= "Going to build for these non-indep arch architectures: ".join(", ", @dscArchitectures).".\n"; }
  } else {
    if ($indepArchitecture eq "all") {
      my @supportedArchitectures = split / /, simplifyList($main::targetArchitectures);
      if ($main::verbosity > 2) { $buildReport .= "Only 'all', going to build for first of my targetArchitectures (".$main::targetArchitectures." -> ".$supportedArchitectures[0].").\n"; }
      push @dscArchitectures, $supportedArchitectures[0];
    } else {
      chdir $workDir || cleanupAndDie("FATAL: Cannot chdir to ".$workDir."\n");
    }
  }

  my @resultFilesToPublish = ();
  #my @resultFilesToTar = ();
  my @logfilesToAttach = ();

  # clean up eventual former 2tar dir, if one run of pba-builder -A processes multiple dsc files:
  system "/bin/rm -rfv /tmp/pba-builder_".$$."/2tar/";
  checkDirectory("/tmp/pba-builder_".$$."/2tar/");

  # iterate over distributions
  foreach my $buildDistribution (@dscDistributions) {
    
    # patching pba-distrubtion in .dsc file ... does not help
    #my $distroDscFile = $dscFile."_".$buildDistribution.".dsc";
    #printToFile($workDir."/".$distroDscFile, $dscText."Pba-Distribution: ".$buildDistribution."\n");
    ##my $distroDscTxt = $dscText;
    ##$distroDscTxt =~ s/Pba-Distributions: .*/Pba-Distribution: $buildDistribution/g;
    ##printToFile($workDir."/".$distroDscFile, $distroDscTxt);

    # iterate over architectures
    foreach my $buildArchitecture (@dscArchitectures) {
      $buildReport .= "\nBuilding for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
      chdir $workDir || cleanupAndDie("FATAL: Cannot chdir to ".$workDir."\n");

      # copy pbuilder
      my $pbuilderTgz = "/var/lib/pba-builder/pbuilder/".$buildDistribution."-".$buildArchitecture."-pba-base.tgz";
      if ($dscBuildDistribution eq "") {
        if ($buildDistribution eq "any") {
          $buildReport .= "\nActually using default distribution ".$main::defaultDistribution." to build for pseudo distro any.\n";
          $pbuilderTgz = "/var/lib/pba-builder/pbuilder/".$main::defaultDistribution."-".$buildArchitecture."-pba-base.tgz";
        }
      } else {
        $buildReport .= "\nActually using XPBA-Build-Distribution ".$dscBuildDistribution." to build for distribution ".$buildDistribution."\n";
        $pbuilderTgz = "/var/lib/pba-builder/pbuilder/".$dscBuildDistribution."-".$buildArchitecture."-pba-base.tgz";
      }
      if (not -f $pbuilderTgz) {
        $buildReport .= "\nPBuilder tarball missing: ".$pbuilderTgz."\nSKIPPING build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        $buildHeadlines .= "SKIPPED build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        next;
      }

      # feed pbuilder with dependend sources.list.d stuff
      my $buildAptSources = "# pba repository, for pba-cbs, pba-utils\ndeb http://deb.clazzes.org//any any-pba-1.0 main\n";
      # deb-src http://deb.clazzes.org//any any-pba-1.0 main";
      $buildAptSources .= "\n# pytools repository, for python3-pba-utils\ndeb http://deb.clazzes.org//any any-pytools-1 main\n";
      # deb https://deb.clazzes.org//any any-pytools-1 main

      my $dcgAptSource = "deb ".$main::repositoryBaseURIs{$buildAptRepository}."/".$main::distributionFamilies{$buildDistribution}." ".$buildDistribution."-".$dscCategory."-".$dscGeneration." main";
      # test if repository is there already!?
      my $packagesURL = $main::repositoryBaseURIs{$buildAptRepository}."/".$main::distributionFamilies{$buildDistribution}."/dists/".$buildDistribution."-".$dscCategory."-".$dscGeneration."/main/binary-".$buildArchitecture."/Packages";
      my $userAgent = LWP::UserAgent->new;
      my $packagesRequest = HTTP::Request->new(GET => $packagesURL);
      my $packagesResponse = $userAgent->request($packagesRequest);
      if ($packagesResponse->is_success()) { 
        if ($main::verbosity > 1) { $buildReport .= "Successfully test-fetched\n".$packagesURL."\nAdding\n".$dcgAptSource."\nas source to be injected.\n"; }
        $buildAptSources .= "\n".$dcgAptSource."\n";
      } else {
        $buildReport .= "Failed to test-fetch\n".$packagesURL."\nNot adding\n".$dcgAptSource."\nas source to be injected.\nThis is ok for the first package of a new distribution-category-generation combination and common for beta and nightly  builds.";
        $buildAptSources .= "# not yet available:\n# ".$dcgAptSource."\n";
      }

      if ($dscVersion =~ /nightly/) {
        if ($main::verbosity > 0) { $buildReport .= "Detected version containing 'nightly', checking nightly repository too.\n"; }
        my $dcgAptSourceNightly = "deb ".$main::repositoryBaseURIs{$buildAptRepository}."/".$main::distributionFamilies{$buildDistribution}." ".$buildDistribution."-".$dscCategory."-".$dscGeneration."nightly main";
        # test if repository is there already!?
        my $packagesURLNightly = $main::repositoryBaseURIs{$buildAptRepository}."/".$main::distributionFamilies{$buildDistribution}."/dists/".$buildDistribution."-".$dscCategory."-".$dscGeneration."nightly/main/binary-".$buildArchitecture."/Packages";
        #my $userAgent = LWP::UserAgent->new;
        my $packagesRequestNightly = HTTP::Request->new(GET => $packagesURLNightly);
        my $packagesResponseNightly = $userAgent->request($packagesRequestNightly);
        if ($packagesResponseNightly->is_success()) { 
          if ($main::verbosity > 1) { $buildReport .= "Successfully test-fetched\n".$packagesURLNightly."\nAdding\n".$dcgAptSourceNightly."\nas source to be injected.\n"; }
          $buildAptSources .= "\n".$dcgAptSourceNightly."\n";
        } else {
          $buildReport .= "Failed to test-fetch\n".$packagesURLNightly."\nNot adding\n".$dcgAptSourceNightly."\nas source to be injected.\nThis is ok for the first package of a new distribution-category-generation combination and common for beta and nightly builds.";
          $buildAptSources .= "# not yet available:\n# ".$dcgAptSourceNightly."\n";
        }
      }
      if ($dscVersion =~ /beta|nightly/) {
        if ($main::verbosity > 0) { $buildReport .= "Detected version containing 'beta' or 'nightly', checking beta repository too.\n"; }
        my $dcgAptSourceBeta = "deb ".$main::repositoryBaseURIs{$buildAptRepository}."/".$main::distributionFamilies{$buildDistribution}." ".$buildDistribution."-".$dscCategory."-".$dscGeneration."beta main";
        # test if repository is there already!?
        my $packagesURLBeta = $main::repositoryBaseURIs{$buildAptRepository}."/".$main::distributionFamilies{$buildDistribution}."/dists/".$buildDistribution."-".$dscCategory."-".$dscGeneration."beta/main/binary-".$buildArchitecture."/Packages";
        #my $userAgent = LWP::UserAgent->new;
        my $packagesRequestBeta = HTTP::Request->new(GET => $packagesURLBeta);
        my $packagesResponseBeta = $userAgent->request($packagesRequestBeta);
        if ($packagesResponseBeta->is_success()) { 
          if ($main::verbosity > 1) { $buildReport .= "Successfully test-fetched\n".$packagesURLBeta."\nAdding\n".$dcgAptSourceBeta."\nas source to be injected.\n"; }
          $buildAptSources .= "\n".$dcgAptSourceBeta."\n";
        } else {
          $buildReport .= "Failed to test-fetch\n".$packagesURLBeta."\nNot adding\n".$dcgAptSourceBeta."\nas source to be injected.\nThis is ok for the first package of a new distribution-category-generation combination and common for beta and nightly builds.";
          $buildAptSources .= "# not yet available:\n# ".$dcgAptSourceBeta."\n";
        }
      }

      my $patchedDscAptSources = $dscAptSources;
      $patchedDscAptSources =~ s/\$family/$main::distributionFamilies{$buildDistribution}/g;
      $patchedDscAptSources =~ s/\$distribution/$buildDistribution/g;
      $patchedDscAptSources =~ s/\$category/$dscCategory/g;
      $patchedDscAptSources =~ s/\$generation/$dscGeneration/g;
      $patchedDscAptSources =~ s/\$mirror/$main::distributionMirrors{$buildDistribution}/g;
      foreach my $repository (@main::repositories) {
        $patchedDscAptSources =~ s/\$repository:$repository/$main::repositoryBaseURIs{$repository}/g;
      }
      $buildAptSources .= $patchedDscAptSources;

      my $script = "#!/bin/sh\n";
      $script .= "export LANG=\"C\"\n";

      if ($main::configParams{"HttpProxy"} ne "") {
        $script .= "echo 'Acquire { http { Proxy \"".$main::configParams{"HttpProxy"}."\"; } }' >/etc/apt/apt.conf.d/999_proxy_http\n";
      }
      if ($main::configParams{"HttpSProxy"} ne "") {
        $script .= "echo 'Acquire { https { Proxy \"".$main::configParams{"HttpProxy"}."\"; } }' >/etc/apt/apt.conf.d/999_proxy_https\n";
      }

      if (scalar(@dscTrustedKeyURLs) > 0) {
        $script .= "apt-get install -y wget sed\n";
        foreach my $keyURL (@dscTrustedKeyURLs) {
          $script .= "wget --no-check-certificate ".$keyURL." -O - | apt-key add -\n";
        }
      } elsif ($dscVersion =~ /beta|nightly/) {
        $script .= "apt-get install -y sed\n";
      }

      $script .= "/bin/mkdir -pv /etc/apt/sources.list.d\n";
      $script .= "/bin/cat >/etc/apt/sources.list.d/pba-builder.list <<EOF\n";
      $script .= $buildAptSources."\nEOF\n\n";
      $script .= "/bin/rm /var/lib/apt/lists/*Packages* /var/lib/apt/lists/*Release* /var/lib/apt/lists/*Sources* >/dev/null 2>/dev/null\n";
      #$script .= "apt-get clean\n";
      $script .= "apt-get update\n"; # || exit 1\n";
      if ($buildAptSources =~ /https:\/\//) {
        $script .= "apt-get install -y apt-transport-https || exit 2\n";
      }

      #$script .= "echo 'Calling set'\n";
      #$script .= "set\n";
      #$script .= "echo '--- enod of set output ---'\n";

      #$script .= "apt-get -y upgrade || exit 2\n";
      #$script .= "apt-get -y dist-upgrade || exit 3\n";

      $script .= "exit 0\n";

      #$buildReport .= "\nPatching these sources into PBuilder tarball:\n".$buildAptSources."\n";
      #$buildReport .= "\nUsing this script to patch the temporary pbuilder tarball:\n".$script."\n";

      chdir $workDir || cleanupAndDie("FATAL: Cannot chdir to ".$workDir."\n");

      # prepare and build

      my $pbuilderCommand = "/usr/sbin/pbuilder build"; # was --build

      my $pbuilderHooksDir = $workDir."/".$buildDistribution."-".$buildArchitecture."_hooks"; 
      checkDirectory($pbuilderHooksDir);
      my $pbaHooksFile = $pbuilderHooksDir."/D01_pba.sh";
      push @main::filesToCleanup, $pbaHooksFile;
      printToFile( $pbaHooksFile, $script);
      
      my $perDistroScript = "#!/bin/sh\n";
      #$perDistroScript .= 'echo "$0 called with params $*"'."\n";
      #$perDistroScript .= 'echo "My content is:"'."\n";
      #$perDistroScript .= 'cat $0'."\n";
      #$perDistroScript .= 'echo "-- SNIP --"'."\n";
      if ($dscVersion =~ /nightly/) {
        #if ($main::verbosity > 2) { $perDistroScript .= "echo 'XBCS-PBA-Generation before adding nightly ...'\ngrep -i PBA-Generation /build/*/debian/control\n"; }
        $perDistroScript .= "/bin/sed -e 's/^XBCS-PBA-Generation: \\(.*\\)/XBCS-PBA-Generation: \\1nightly/i' -i /build/*/debian/control\n";
        #$perDistroScript .= "/bin/sed -e 's/^\\(XBCS-\\)\\?PBA-Generation: \\(.*\\)/\\1PBA-Generation: \\2nightly/i' -i /build/*/debian/control\n";
        #if ($main::verbosity > 2) { $perDistroScript .= "echo 'XBCS-PBA-Generation after adding nightly ...'\ngrep -i PBA-Generation /build/*/debian/control\n"; }
      } else {
        $perDistroScript .= 'echo "No need to append \'nightly\' to PBA-Generation"'."\n";
        if ($dscVersion =~ /beta/) {
          #if ($main::verbosity > 2) { $perDistroScript .= "echo 'XBCS-PBA-Generation before adding beta ...'\ngrep -i PBA-Generation /build/*/debian/control\n"; }
          $perDistroScript .= "/bin/sed -e 's/^XBCS-PBA-Generation: \\(.*\\)/XBCS-PBA-Generation: \\1beta/i' -i /build/*/debian/control\n";
          #$perDistroScript .= "/bin/sed -e 's/^\\(XBCS-\\)\\?PBA-Generation: \\(.*\\)/\\1PBA-Generation: \\2beta/i' -i /build/*/debian/control\n";
          #if ($main::verbosity > 2) { $perDistroScript .= "echo 'XBCS-PBA-Generation after adding beta ...'\ngrep -i PBA-Generation /build/*/debian/control\n"; }
        } else {
          $perDistroScript .= 'echo "No need to append \'beta\' to PBA-Generation"'."\n";
        }
      }
      $perDistroScript .= "/bin/egrep -i '^XBCS-PBA-Distribution: ' /build/*/debian/control && exit 0\n";
      $perDistroScript .= "/bin/sed -e 's/^XBCS-PBA-Distributions: .*/XBCS-PBA-Distribution: ".$buildDistribution."/' -i /build/*/debian/control\n";
      #$perDistroScript .= "echo 'Searching /build'\n";
      #$perDistroScript .= "find /build -type f\n";
      #$perDistroScript .= 'echo "-- SNIP --"'."\n";
      #if ($main::verbosity > 2) { $buildReport .= "\nA01_pba.sh for ".$buildDistribution." looks like this:\n".$perDistroScript."\n--- SNIP ---\n\n"; }
      my $pbaPerDistroHooksFile = $pbuilderHooksDir."/A01_pba.sh";
      push @main::filesToCleanup, $pbaPerDistroHooksFile;
      printToFile( $pbaPerDistroHooksFile, $perDistroScript);
      
      $pbuilderCommand .= " --hookdir ".$pbuilderHooksDir;

      if ($main::pbabuilderArchitecture ne $buildArchitecture) {
        $pbuilderCommand .= " --host-arch ".$buildArchitecture;
        if (-f "/etc/pbuilderrc.cross") {
          $pbuilderCommand .= " --configfile /etc/pbuilderrc.cross";
        }
      }
      $pbuilderCommand .= " --basetgz ".$pbuilderTgz;
      if ($main::configParams{"HttpProxy"} ne "") {
        $pbuilderCommand .= " --http-proxy ".$main::configParams{"HttpProxy"};
      }
      my $pbuilderBuildLog = $workDir."/build_".$buildDistribution."-".$buildArchitecture.".log";
      $pbuilderCommand .= " --logfile ".$pbuilderBuildLog;
      my $pbuilderResultsDir = $workDir."/".$buildDistribution."-".$buildArchitecture;
      checkDirectory($pbuilderResultsDir);
      $pbuilderCommand .= " --buildresult ".$pbuilderResultsDir;

      # signing options
      $pbuilderCommand .= ' --debbuildopts "'.$main::dpkgGpgOptions.'"';

      # FIXME add --hook-source=foobar.sh to change 
      #   Pba-Distributions: wheezy jessie
      # to
      #   Pba-Distribution: $buildDistribution
      # in debian/control
      
      # FIXME add --hook-init=foobar.sh to do what we now do with tarball patching

      $pbuilderCommand .= " ".$dscFile; #$distroDscFile;

      push @logfilesToAttach, $pbuilderBuildLog;
      if ($main::verbosity > 0) { $buildReport .= "Executing pbuilder build in ".pwd().":\n".$pbuilderCommand."\n"; }
      my $pbuilderBuildRc = system($pbuilderCommand);
      if ($pbuilderBuildRc eq 0) {
        $buildReport .= "\nSUCESSFULLY built for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        $buildReport .= "See attached build_".$buildDistribution."-".$buildArchitecture.".log\n\n";
        $buildHeadlines .= "SUCCESSFULLY built for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
      } else {
        $buildReport .= "\nFAILED build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.", pbuilder exit code: ".$pbuilderBuildRc.".\n";
        $buildReport .= "See attached build_".$buildDistribution."-".$buildArchitecture.".log\n\n";
        $buildHeadlines .= "FAILED to build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.", pbuilder exit code: ".$pbuilderBuildRc.".\n";
      }

      # /bin/rm patched files (in any case!)
      #if ($main::verbosity <= 2) {
        #unlink $dscFile.".ori", $changesFile.".ori";
        #unlink $patchedDscFile, $patchedChangesFile;
      #}

      $buildReport .= "Build done for distribution ".$buildDistribution.", architecture ".$buildArchitecture.":\n";

      # mv .deb files to 2tar/ dir, remove other files
      chdir $pbuilderResultsDir || cleanupAndDie("FATAL: Cannot chdir to ".$pbuilderResultsDir."\n");
      my @producedFiles = glob("*");
      my $debFileSearch = "(_[a-z0-9]+).deb".'$';
      #my $debFileReplace = "_".$buildDistribution.'\\1.deb';
      foreach my $producedFile (@producedFiles) {
        next if (! -f $producedFile); # skip dirs ...
        if ($producedFile =~ /$debFileSearch/) {
          #my $patchedFilename = $producedFile;
          #$patchedFilename =~ s/$debFileSearch/_$buildDistribution$&/;
          checkDirectory("/tmp/pba-builder_".$$."/2tar/".$buildDistribution);
          executeOrDie("/bin/mv -v ".$pbuilderResultsDir."/".$producedFile." /tmp/pba-builder_".$$."/2tar/".$buildDistribution."/".$producedFile);
          #push @resultFilesToTar, $buildDistribution."/".$producedFile;
          $buildReport .= "/tmp/pba-builder_".$$."/2tar/".$buildDistribution."/".$producedFile."\n";
        } else {
          unlink $producedFile;
        }
      }
      $buildReport .= "\n";
      chdir "..";
      executeOrDie ("/bin/rm -rf ".$pbuilderResultsDir);
    }
    #executeOrDie ("/bin/cp -uva ".$workDir."/".$distroDscFile." /tmp/");
    #executeOrDie ("/bin/rm -rfv ".$workDir."/".$distroDscFile);
  }

  # tar stuff
  my $tarLog = $workDir."/tar_".$dscSrcName."_".$dscVersion.".log";
  push @logfilesToAttach, $tarLog;
  my $resultTarball = $main::resultsDir."/".$dscSrcName."_".$dscVersion.".debs.tar.gz";
  chdir "/tmp/pba-builder_".$$."/2tar" || cleanupAndDie("FATAL: Cannot chdir to /tmp/pba-builder_".$$."/2tar.\n");
  my $tarCmd = "tar cvzf ".$resultTarball." . 2>&1 |tee ".$tarLog;
  if (not execute($tarCmd)) {
    $buildReport .= "\nFATAL: Unable to collect tarball, see attached ".$tarLog."\n";
    $buildHeadlines .= "FAILED to collect tarbarball.\n";
    ($main::verbosity > 0) && print $buildReport."\n".$buildHeadlines;
    report ( $buildHeadlines."\n\n".$buildReport, $maintainers, join(" ", @logfilesToAttach) );
    return;
  }
  $buildReport .= "\nSucessfully collected tarball ".$tarLog."\n";
  $buildHeadlines .= "SUCCESSFULLY collected tarbarball.\n";
  push @resultFilesToPublish, $resultTarball; 
  chdir $main::resultsDir;

  # construct $feedScripts
  my @feedScriptsToCall = ();
  my @feedScriptsToPropose = ();
  
  my @uniqFilesToPublish = uniq(@resultFilesToPublish);

  if (scalar(@uniqFilesToPublish) == 0) {

    $buildReport .= "No resulting files to feed a repository with.\n\n";
    $buildHeadlines .= "NO resulting files to feed a repository with.\n\n";

    if ($main::localTargetDirectory ne "") {
      if (scalar(@logfilesToAttach) > 0) {
        executeOrDie("/bin/cp -av ".join(" ", @logfilesToAttach)." ".$main::localTargetDirectory."/");
      }
    }

  } else {

    my $uploadName = $dscFile;
    $uploadName =~ s/\.dsc$/_debs/;

    if ($main::localTargetDirectory ne "") {

      my $feedScriptName = "feed_".$uploadName."_to_localdir.sh";
      my $feedScript = proposeMvScript( $main::localTargetDirectory, join(" ", @uniqFilesToPublish) );
      if (scalar(@logfilesToAttach) > 0) {
        $feedScript .= "/bin/cp -av ".join(" ", @logfilesToAttach)." ".$main::localTargetDirectory."/";
      }
      $feedScript .= "/bin/cp -av /tmp/pba-builder_".$$."/2tar/* ".$main::localTargetDirectory."/";
      $feedScript .= "\nexit 0\n";
      printToFile ($main::resultsDir."/".$feedScriptName, $feedScript, 0755);
      push @feedScriptsToCall, $main::resultsDir."/".$feedScriptName;

    } else {
 
      foreach my $dscRepository (@dscRepositories) {
        if ($main::repositoryFeedHostDirs{$dscRepository} eq "") {
          $buildReport .= "I don't know how to feed this repository:\n# ".$main::repositoryLongNames{$dscRepository}."\n\n";
          $buildHeadlines .= "SKIPPED feeding repository ".$dscRepository."\n";
          next;
        }
        my ($feedHost, $feedDir) = split /:/, $main::repositoryFeedHostDirs{$dscRepository}, 2;
        my $uploadDir = $feedDir."/uploading-".$uploadName."_".$$;
        my $feedScriptName = "feed_".$uploadName."_to_".$dscRepository.".sh";
        my $feedScript = "#!/bin/sh\n";
        $feedScript = "cd ".$main::resultsDir." || exit 1\n";
        # create remote upload dir
        $feedScript .= "echo 'Calling ...\n/usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir."'\n";
        $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir." \\\n";
        $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir." \\\n";
        $feedScript .= "  || exit 2\n";
        $feedScript .= "echo 'Done.\n'\n";
        # echo /usr/bin/scp files
        $feedScript .= "echo 'Calling ...\n/usr/bin/scp \\\n";
        foreach my $fileToPublish (@uniqFilesToPublish) {
          $feedScript .= "  ".$fileToPublish." \\\n";
        }
        $feedScript .= "  ".$feedHost.":".$uploadDir."/'\n";
        # /usr/bin/scp files
        $feedScript .= "/usr/bin/scp \\\n";
        foreach my $fileToPublish (@uniqFilesToPublish) {
          $feedScript .= "  ".$fileToPublish." \\\n";
        }
        $feedScript .= "  ".$feedHost.":".$uploadDir."/ \\\n";
        $feedScript .= "  || /usr/bin/scp \\\n";
        foreach my $fileToPublish (@uniqFilesToPublish) {
          $feedScript .= "  ".$fileToPublish." \\\n";
        }
        $feedScript .= "  ".$feedHost.":".$uploadDir."/ \\\n";
        $feedScript .= "  || exit 3\n";
        $feedScript .= "echo 'Done.\n'\n";
        # /bin/mv files
        $feedScript .= "echo 'Calling ...\n/usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/'\n";
        $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/ \\\n";
        $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/ \\\n";
        $feedScript .= "  || exit 4\n";
        $feedScript .= "echo 'Done.\n'\n";
        # /bin/rm remote uploading dir
        $feedScript .= "echo 'Calling ...\n/usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir."'\n";
        $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir." \\\n";
        $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir." \\\n";
        $feedScript .= "  || echo 'WARNING: Failed to rm empty upload dir on repository host.'\n";
        #$feedScript .= "  || exit 5\n";
        $feedScript .= "echo 'Done.\n'\n";
        # /bin/rm files if we are the last feed script
        $feedScript .= "# if we are the only feed script left, remove files\n";
        $feedScript .= 'OTHERFEEDERS=$(ls -1 |egrep "feed_'.$uploadName.'.*\.sh$" |egrep -v '.$feedScriptName.")\n";
        $feedScript .= 'if [ -z "$OTHERFEEDERS" ] ; then'."\n";
        # echo /bin/rm
        $feedScript .= "  echo 'Calling ...\n/bin/rm -v \\\n";
        foreach my $fileToPublish (@uniqFilesToPublish) {
          $feedScript .= "    ".$fileToPublish." \\\n";
        }
        $feedScript .= "'\n";
        # rm
        $feedScript .= "  /bin/rm -v \\\n";
        foreach my $fileToPublish (@uniqFilesToPublish) {
          $feedScript .= "    ".$fileToPublish." \\\n";
        }
        $feedScript .= "    || exit 6\n";
        $feedScript .= "  echo 'Done.\n'\n";
        $feedScript .= "else\n";
        $feedScript .= "  echo 'Not removing ... \n";
        foreach my $fileToPublish (@uniqFilesToPublish) {
          $feedScript .= "    ".$fileToPublish."\n";
        }
        $feedScript .= "... because I found the following other feeders:'\n";
        $feedScript .= '  echo "$OTHERFEEDERS"'."\n";
        $feedScript .= "fi\n";
        #$feedScript .= "# finally, suicide\n";
        #$feedScript .= "/bin/rm -fv ".$feedScriptName."\n";
        $feedScript .= "exit 0\n";
        if ($main::repositoryAutoFeed{$dscRepository} eq "true") {
          push @feedScriptsToCall, $main::resultsDir."/".$feedScriptName;
        } else {
          push @feedScriptsToPropose, $main::resultsDir."/".$feedScriptName;
        }
        printToFile ($main::resultsDir."/".$feedScriptName, $feedScript, 0755);
      }
    }

    if (scalar(@feedScriptsToCall) > 0) {
      foreach my $feedScript (@feedScriptsToCall) {
        my $feedScriptBase = $feedScript;
        $feedScriptBase =~ s/^.*\///g;
        my $feedLog = $workDir."/".$feedScriptBase.".log";
        push @logfilesToAttach, $feedLog;
        my $feedRc = system("/bin/sh ".$feedScript." >".$feedLog." 2>&1");
        if ($feedRc == 0) {
          $buildReport .= "Successfully executed this feed script:\n".$feedScript."\n";
          $buildHeadlines .= "SUCCESSFULLY executed ".$feedScript."\n";
          if (-f $feedScript) { unlink $feedScript; }
        } else {
          $buildReport .= "ERROR: Failed executing this feed script (rc=".$feedRc."):\n".$feedScript."\n";
          $buildHeadlines .= "FAILED to execute ".$feedScript."\n";
        }
      }
    }

    if (scalar(@feedScriptsToPropose) > 0) {
      $buildReport .= "\nProposing to call these feed-and-cleanup scripts:\n";
      $buildHeadlines .= "\nTo publish, PLEASE CALL these feed-and-cleanup scripts:\n";
      foreach my $feedScript (@feedScriptsToPropose) {
        $buildReport .= $feedScript." && \\\n  /bin/rm -fv ".$feedScript."\n";
        $buildHeadlines .= $feedScript." && \\\n  /bin/rm -fv ".$feedScript."\n";
      }
    }


  } # end if no resultFilesToPublish
 
  ($main::verbosity > 0) && print $buildReport."\n".$buildHeadlines;
  report ( $buildHeadlines."\n\n".$buildReport, $maintainers, join(" ", @logfilesToAttach) );

  #if ($main::verbosity <= 2) {
 
    unlink @logfilesToAttach;
 
    # /bin/rm main dsc and childs (in any case!)
    chdir $workDir || cleanupAndDie("FATAL: Cannot chdir to ".$workDir."\n");
    unlink $dscFile || cleanupAndDie("FATAL: Cannot /bin/rm ".pwd()."/".$dscFile."\n");
    for my $childFile (@dscChildFiles) {
      unlink $childFile || cleanupAndDie("FATAL: Cannot /bin/rm ".pwd()."/".$childFile."\n");
    }
    (-f $changesFile) && unlink $changesFile; 
    chdir "..";
    executeOrDie("/bin/rm -rfv ".$workDir);

  #}

  ($main::verbosity > 1) && print "Done building ".$dscFile."\n";
}



sub executeInPbuilder( $ $ $ $ ) {
  # executes a script "in" a pbuilder
  # $1 = script text (including apropriate line 1 like #!/bin/sh) or local file 
  # $2 = pbuilder tgz
  # $3 = logfile, may be empty string
  # $4 = architecture
  # dies if parameters are empty
  # returns ...
  my $script = shift;
  ($script eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty script string.\n");
  my $pbuilderTgz = shift;
  ($pbuilderTgz eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty tarball parameter.\n");
  (-f $pbuilderTgz) || cleanupAndDie("FATAL: executeInPbuilder cannot find pbuilder tarball ".$pbuilderTgz."\n");
  my $pbuilderLog = shift;
  # may be empty
  my $architecture = shift;
  ($architecture eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty architecture.\n");

  my $scriptFile;
  my $fileToCleanup = "";
  if ( ($script !~ /\n/) and ($script !~ /\//) and (-x $script) ) {
    #($main::verbosity > 2) && print "executeInPbuilder: Interpretating \$1 (".$scriptFile.") as script, not as set of commands.\n"; 
    $scriptFile = $script;
  } else {
    chdir "/tmp" || cleanupAndDie("FATAL: Cannot chdir to /tmp\n");
    $scriptFile = $$."_pba4pbuilder.sh";
    $fileToCleanup = $scriptFile;
    #($main::verbosity > 2) && print "executeInPbuilder: Interpretating \$1 as script text, creating ".$scriptFile.".\n"; 
    if (-f $scriptFile) { unlink $scriptFile || cleanupAndDie("FATAL: Cannot delete obsolete ".$scriptFile."\n"); }
    printToFile( $scriptFile, $script, 0755 );
    push @main::filesToCleanup, $scriptFile;
  }

  my $scriptOptions = "";

  my $pbuilderCommand = "/usr/sbin/pbuilder execute"; # was --execute
  # should not hurt:
  if ($main::pbabuilderArchitecture ne $main::targetArchitecture) {
    $pbuilderCommand .= " --host-arch ".$main::targetArchitecture;
  }
  $pbuilderCommand .= " --basetgz ".$pbuilderTgz;
  $pbuilderCommand .= " --save-after-exec";
  if ($main::configParams{"HttpProxy"} ne "") {
    $pbuilderCommand .= " --http-proxy ".$main::configParams{"HttpProxy"};
  }

  if ($pbuilderLog eq "") {
    $pbuilderCommand .= " --logfile /var/log/pbuilder/executeInPbuilder_".$$.".log";
  } else {
    $pbuilderCommand .= " --logfile ".$pbuilderLog;
  }

  $pbuilderCommand .= " -- ".$scriptFile;
  if ($scriptOptions ne "") {
    $pbuilderCommand .= " ".$scriptOptions;
  }

  ($main::verbosity > 0) && print "Executing this pbuilder command:\n".$pbuilderCommand."\n";
  dieIfNotRoot();

  my $rc = system($pbuilderCommand);

  if ( ($fileToCleanup ne "") and (-f $fileToCleanup) )
  { unlink $fileToCleanup || cleanupAndDie("FATAL: Cannot delete obsolete ".$scriptFile."\n"); }

  if ($rc == 0) {
    ($main::verbosity > 1) && print "pbuilder successful, see ".$pbuilderLog."\n";
    return 1;
  } else {
    ($main::verbosity > 1) && print "ERROR: pbuilder failed (rc=".$rc."), see ".$pbuilderLog."\n";
    return 0;
  }

}



1;

